//------------------------------------------------------------------------
//
// OpenDCC - DCC Packet Receiver Module
//
// Copyright (c) 2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      TimerPWM.c
// author:    Wolfgang Kufer
// webpage:   http://opendcc.sourceforge.net/
// history:   2008-01-23 as Initial Version based on Wolfgang Kufer's
//                          dcc_receiver.[hc] files 
//            2008-04-14 kw added code for Sniffer
//                          This has only one capture on timer 0; so I used
//                          Timer 2 to do the system clock
//------------------------------------------------------------------------
// credit:    This software is substantially based on work by
//            Wolfgang Kufer (kufer@gmx.de) His original project
//            can be found here: http://www.opendcc.de
//------------------------------------------------------------------------
//
// purpose:   This library module implement a DCC Packet Receiver that uses
//            an External Interrupt line INTx and a timer to decode DCC packets
//
//------------------------------------------------------------------------

#include <inttypes.h>
#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "TimerPWM.h"

//---------------------------------------------------------------------------
// Define TIMER2 handling macros
//
// Define the TIMER2, this is a 8 Bit Timer, Overflow at 256
//
//
// we run Timer2 with 16MHz, Prescaler 8: This is 0.5us grid; 128us Systimer
//
#define SYSTIMER_PERIOD  128L 
#define T2_PRESCALER   8 

volatile signed char systick_128us;    // gets incremented every 128us
volatile signed char systick_1ms;      // gets incremented every 1ms

unsigned long measure_time;            // gets incremented every 1ms
//---------------------------------------------------------------------------

#if   (T2_PRESCALER==1)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==8)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==32)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==64)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==128)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==256)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==1024)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(1<<CS20))
#else
  #error T2_PRESCALER is not defined or is not a supported value
#endif

void init_timer(void)
  {
    TCCR2 = (0<<FOC2)   // reserved set to 0
          | (1<<WGM20)  // Mode 3: Fast PWM (Top = 0xFF)
          | (1<<COM21)  // 1,0 Clear OC2 on Compare Match, set OC2 at TOP.
          | (0<<COM20)
          | (1<<WGM21)
          | T2_PRESCALER_BITS;

    TCNT2 = 0;

    TIMSK |= (1<<TOIE2);  // enable overflow interrupt

    systick_128us = 0;
    systick_1ms = 0;
    measure_time = 0;
  }

void set_pwm_value(unsigned char pwm)
  {
    OCR2 = pwm;
  }



ISR(TIMER2_OVF_vect)
  {
    // We get an Interrupt every 128us
    // We do a software counter - fast_systick gets incremented every int;
    // 
    systick_128us++;
  }  

signed char local_systick;
signed char local_meastick;

void run_timer(void)
  {
    signed char compare;

    compare = systick_128us - local_systick;
    if (compare >= 8)
      {
        systick_1ms++;
        local_systick += 8;
      }

    compare = systick_128us - local_meastick;
    if (compare >= 8)
      { // 1.024ms
        measure_time++;
        if (measure_time >= 1000000L) measure_time = 0;
        local_meastick += 8;
      }
 
  }

void reset_measure_timer(void)
  {
    measure_time = 0;
    local_meastick = systick_128us;  // reset fractional part
  }

//--------------------------------------------------------------------
// tachometer
//--------------------------------------------------------------------
//
// we do rotational speed measurements with timer1
//
// measurement range:  1rps - 400rps (= 60rpm - 24000rpm)
//
// method:  we use the capture interrupt to get timestamps from sensor events
//          at the shaft. The timer runs at 16MHz / 1024 = 64us clock rate.
// 
// sensor:  we use an IR-reflective sensor. The sensor output is compared
//          with the analog comparator against a variable threshold. This
//          threshold is created with the pwm-output of timer2 and an
//          external lowpass.
//
// values:  speed:     Interrupts rate:      counter val:    average
//          400rps        2.5ms                   39              6
//          200rps        5  ms                   78              3
//          100rps       10  ms                  156              2
//           50rps       20  ms                  312              1
//           10rps      100  ms                 1562              1
//            1rps     1000  ms                15625              1
//
// flow:    sensor -> ISR -> copy counter value to circular buffer.
//                        -> set event flag.
//
//          systick -> if event:
//                        -> read buffer ... buffer(average).
//                        -> calc rpm
//                     else
//                        -> check timeout
//
// Note:    to get better measurement resolution a low speed, it is recommended
//          to use two or more reflectors.
//

volatile uint16_t buffer[8];
volatile uint16_t last_capture;
volatile uint8_t buf_index;
volatile uint8_t capture_event;

ISR(TIMER1_CAPT_vect)
  {
    int16_t capture;
    uint8_t my_index;

    capture = ICR1;

    my_index = buf_index + 1;       // most efficient code
    my_index &= 0x07;
    buf_index = my_index;

    // my_index = buf_index + 1;
    // if (my_index >= 8) my_index = 0;
    // buf_index = my_index;



    // buf_index++;
    // if (buf_index >= 8) buf_index = 0;


    buffer[buf_index] = capture - last_capture;

    last_capture = capture;
  }


void evaluate(void)
  {
    
    cli();

  }
